#include "functions.h"

/*Macro que dada una estructura tivemal de un paquete pcap devuelve su tiempo en us*/
#define GETUS(t) ((t.tv_sec*1000000)+(t.tv_usec))


/*Tabla con pares puerto valor utilizada para implementar los top de puertos*/
top_ports ports[MAX_PORTS];



/*********************************************************************************************************************************************************
 * Función:createCDF
 * Descripción: Esta función crea un fichero que contiene los puntos necesarios para dibujar la ECDF dados los datos contenidos en el fichero filename_data.
 * Entrada:
 *
 *	-char *filename_data: nombre del fichero de entrada con los valores de las muestras
 *	-char *filename_cdf: nombre del fichero de salida con los valores para dibujar la ECDF
 *	
 * Salida:
 *	- 0 si no hay ningún problema
 *	- -1 si se produce algún error.
 *
 *  Autor: Jose Luis Garcia Dorado
 *
 *  (C) 2013 EPS-UAM 
 ****************************************************************************************************************************************************************/
int createCDF(char* filename_data, char* filename_cdf) {
	char comando[255]; char linea[255]; char aux[255];
	char *aux_file="aux_file";
	int num_lines;
	FILE *f;

	sprintf(comando,"wc -l %s 2>&1",filename_data);
	f = popen(comando, "r");
	if(f == NULL){
		printf("Error ejecutando el comando\n");
		return ERROR;
	}
	fgets(linea,255,f);
	sscanf(linea,"%d %s",&num_lines,aux);
	pclose(f);

	sprintf(comando,"sort -n < %s > %s 2>&1",filename_data,aux_file);
	f = popen(comando, "r");
	if(f == NULL){
		printf("Error ejecutando el comando\n");
		return ERROR;
	}
	bzero(linea,255);
        fgets(linea,255,f);
	pclose(f);

	/********************************************************************************************
	crear CDF [...]
	para ello abrir el fichero aux_file y leer cada línea. Para cada línea calcular
	su valor de probabilidad acumulada. Escribir en filename_cdf el valor de la muestra junto con 
	su valor de probabilidad acumulada. Las dos magnitudes deben estar separadas por una coma
	Ej: 400,0.578
	*********************************************************************************************/
	

	return OK;
}



/*********************************************************************************************************************************************************
 * Función:getStats

 * Descripción: Esta función abre un fichero pcap y realiza el análisis de bytes,paquetes y duración en función de los filtros especificados como entrada.
 * Entrada:
 *	-uint32_t *bytes: Contador de bytes analizados. En caso de que haya filtros de 
 *	cualquier tipo (BPF, puertos, IPs) este valor debe almacenar el número de bytes 
 *	recibidos tras aplicar el filtro. Este valor debe rellenarse dentro de la función
 *	 
 *	 -uint32_t *packets: Contador de paquetes analizados. En caso de que haya filtros de 
 *	cualquier tipo (BPF, puertos, IPs) este valor debe almacenar el número de paquetes 
 *	recibidos tras aplicar el filtro. Este valor debe rellenarse dentro de la función
 *
 *	 -uint64_t *duration: Duración del tráfico analizado expresado en us. En caso de que haya filtros de 
 *	 cualquier tipo (BPF, puertos, IPs) este valor debe almacenar la duración de tráfico
 *	 tras aplicar el filtro. Este valor debe rellenarse dentro de la función
 *
 *	-uint16_t ethertype: Ethertype utilizado para filtrar. En caso de que se pase un valor diferente a 0xFFFF sólo deben
 *	analizarse aquellos paquetes que contengan el ethertype especificado.
 *	
 *	-uint32_t ip_src: Dirección IP origen utilizada para filtrar. En caso de que se pase un valor diferente a 0xFFFFFFFF sólo deben
 *	analizarse aquellos paquetes que contengan la ip origen especificada.
 *	
 *	-uint32_t ip_dst: Dirección IP destino utilizada para filtrar. En caso de que se pase un valor diferente a 0xFFFFFFFF sólo deben
 *	analizarse aquellos paquetes que contengan la ip destino especificada.
 *
 *	-uint16_t port_src: Puerto origen utilizado para filtrar (sólo aplica a paquetes TCP/UDP). En caso de que se pase un valor diferente a 0xFFFF sólo deben
 *	analizarse aquellos paquetes que contengan el puerto origen especificado.
 *	
 *	-uint16_t port_dst: Puerto destino utilizado para filtrar (sólo aplica a paquetes TCP/UDP). En caso de que se pase un valor diferente a 0xFFFF sólo deben
 *	analizarse aquellos paquetes que contengan el puerto destino especificado.
 *
 *	-uint8_t proto: Protocolo (a nivel IP) utilizado para filtrar. En caso de que se pase un valor diferente a 0xFF sólo deben
 *	analizarse aquellos paquetes que contengan el protocolo especificado.
 *
 *	-char *filter: En caso de que se pase un valor distinto de NULL debe aplicarse el filtro BPF contenido en filter al realizar el análisis del tráfico.
 *
 *	-char *filename: Nombre del fichero pcap a analizar
 *	
 * Salida:
 *	- 0 si no hay ningún problema
 *	- -1 si el nombre de fichero pasado es NULL o se produce algún error al procesar el fichero pcap especificado.
 *
 *  Autor: Javier Ramos de Santiago
 *
 *  (C) 2013 EPS-UAM 
 ****************************************************************************************************************************************************************/


int getStats(uint32_t *bytes,uint32_t *packets,uint64_t *duration,uint16_t ethertype,uint32_t ip_src, uint32_t ip_dst, uint16_t port_src, uint16_t port_dst, uint8_t proto, char * filter, char* filename)
{
	
	pcap_t *handle;
	char errbuf[PCAP_ERRBUF_SIZE];
	struct pcap_pkthdr header;
	int flag=0;
	const u_char* packetsaux;
	uint64_t ts_last,ts_first;
	uint8_t ihlaux=0;
	uint16_t portorig;
	uint16_t portdest;
	uint16_t etheraux=0;
	uint8_t sizeRead=0;	
	uint8_t ihl=0,longaux=0;
	uint32_t ip_srcaux,ip_destaux;
	uint8_t protocolo=0;
	handle=pcap_open_offline(filename,errbuf);
	if(handle==NULL){
		fprintf(stderr,"No se puede abrir traza.pcap %s\n",errbuf);
		return -1;
	}	
      while((packetsaux=pcap_next(handle,&header))!=NULL){	

		if(ethertype==NO_FILTER_ETHER && ip_src==NO_FILTER_IP && ip_dst==NO_FILTER_IP && port_src==NO_FILTER_PORT && port_src==NO_FILTER_PORT && 				  			proto ==NO_FILTER_PROTO){
		 (*packets)++;		
		(*bytes)+=header.len;
	    	if(flag==0){
			ts_first=GETUS(header.ts);
			flag=1;
		}
		}	
			else{ 
	
        sizeRead=12;//saltamos 12 bytes para copiar el tipo ethernet       
        memcpy(&etheraux,packetsaux+sizeRead,sizeof(uint16_t));
		sizeRead+=sizeof(uint16_t);
		etheraux=ntohs(etheraux);// TIPO ETHERNET

       if((ethertype!=0xFFFF)&&(etheraux!=ethertype))
			continue;

		//TIPO IPv4
	   if (etheraux==0x0800){
		longaux=0;
		longaux=sizeRead;
		memcpy(&ihlaux,sizeRead+packetsaux,sizeof(uint8_t));//IHL
		sizeRead+=sizeof(uint8_t);

		ihl=ihlaux&(0x0F); //mascara para sacar el ihl 
		ihl=ihl*4;				
		sizeRead+=8*sizeof(uint8_t);

		memcpy(&protocolo,sizeRead+packetsaux,sizeof(uint8_t));//PROTOCOLO
		sizeRead+=sizeof(uint8_t);
		if((proto!=0xFF)&&(protocolo!=proto) )
			continue;
		  

			sizeRead+=sizeof(uint16_t); //leemos la ip origen
			memcpy(&ip_srcaux,sizeRead+packetsaux,sizeof(uint32_t));
			sizeRead+=sizeof(uint32_t);
			ip_srcaux=ntohl(ip_srcaux); 
 			if((ip_src!=0xFFFFFFFF)&&(ip_srcaux!=ip_src)) 
				continue;
	
			memcpy(&ip_destaux,sizeRead+packetsaux,sizeof(uint32_t)); //leemos la ip destino
			sizeRead+=sizeof(uint32_t);
			ip_destaux=ntohl(ip_destaux);
			if((ip_dst!=0xFFFFFFFF)&&(ip_destaux!=ip_dst)) 
				continue;			
			sizeRead=longaux+ihl;			
			if((protocolo==6)||(protocolo==17))	{    //si el protocolo es tcp o UDP seguimos leyendo
		 			memcpy(&portorig,sizeRead+packetsaux,sizeof(uint16_t));
				    sizeRead+=sizeof(uint16_t);
					portorig=ntohs(portorig); //lectura del puerto origen
					if((port_src!=0xFFFF)&&(portorig!=port_src)) 
							continue;
		
		 			memcpy(&portdest,sizeRead+packetsaux,sizeof(uint16_t)); //lectura del puerto destino
		 			sizeRead+=sizeof(uint16_t);
					portdest=ntohs(portdest);
	      			if((port_dst!=0xFFFF)&&(portdest!=port_dst)) 
							continue;

					(*packets)++;
					(*bytes)+=header.len;
			 		if(flag==0){
						ts_first=GETUS(header.ts);
						flag=1;
			 		}			
	     		    
	      }   
			else{
					(*packets)++;
					(*bytes)+=header.len;
			 		if(flag==0){
						ts_first=GETUS(header.ts);
						flag=1;
			 		}			

				}   
		
	}
	else
	{
	
			(*packets)++;
			(*bytes)+=header.len;
			 if(flag==0){
				ts_first=GETUS(header.ts);
				flag=1;
			 }			
	}
	


	};
}

     
    	
	ts_last=GETUS(header.ts);
	*duration=ts_last-ts_first;
     pcap_close(handle);   
	
	return 0;
}


/*********************************************************************************************************************************************************
 * Función:getTopNPort
 * Descripción: Esta función devuelve una lista de los N puertos (origen o destino) más populares agrupados por bytes o paquetes en función 
 *		de los parámetros pasados como argumento. 
 * Entrada:
 *
 *	-uint8_t direction: En caso de que la variable contenga el valor 0 se deben analizar los puertos origen. En caso de que contenga 1 se deben analizar los puertos destino.
 *	-uint8_t bytes_packets: En caso de que la variable contenga el valor 0 se debe contabilizar el número de bytes por puerto para realizar el Top. 
 *				 En caso de que contenga 1 se debe contabilizar el número de paquetes por puerto para realizar el Top. 
 *	-uint8_t N: Número de resultados a mostrar en el Top.
 *	-uint16_t *port: Array de tamaño N en el que deben guardarse los N puertos (origen o destino) más populares (por bytes o paquetes). Se pasa reservado.
 *	-uint16_t *value: Array de tamaño N en el que deben guardarse los valores (de bytes o paquetes) observados en los N puertos (origen o destino) más populares. Se pasa reservado
 *	-char *filter: En caso de que se pase un valor distinto de NULL debe aplicarse el filtro BPF contenido en filter al realizar el análisis de puertos.
 *
 *	-char *filename: Nombre del fichero pcap a analizar
 *	
 * Salida:
 *	- 0 si no hay ningún problema
 *	- -1 si el nombre de fichero pasado es NULL o se produce algún error al procesar el fichero pcap especificado.
 *
 *  Autor: Javier Ramos de Santiago
 *
 *  (C) 2013 EPS-UAM 
 ****************************************************************************************************************************************************************/

int getTopNPort(uint8_t direction,uint8_t bytes_packets,uint8_t N, uint16_t *port, uint32_t *value,u_char * filter, char* filename)
{

	bzero(&ports,MAX_PORTS*sizeof(top_ports));





















	return 0;


}
/*********************************************************************************************************************************************************
 * Función:getSizeECDFPort
 * Descripción: Esta función abre un fichero pcap y  escribe en el fichero apuntado por filename_out los valores correspondiente para la ECDF por tamaños 
 *filtrando por los puertos origen y destino que se pasan en port_src y port_dst.
 * Entrada:
 *	-char *filename: Nombre del fichero pcap a analizar
 *	-char *filename_out: Nombre del fichero de salida donde se volcaran los valores de la ECDF
 *	-uint16_t port_src: Puerto origen utilizado para filtrar (sólo aplica a paquetes TCP/UDP). En caso de que se pase un valor diferente a 0xFFFF sólo deben
 *	analizarse aquellos paquetes que contengan el puerto origen especificado.
 *	
 *	-uint16_t port_dst: Puerto destino utilizado para filtrar (sólo aplica a paquetes TCP/UDP). En caso de que se pase un valor diferente a 0xFFFF sólo deben
 *	analizarse aquellos paquetes que contengan el puerto destino especificado.
 *
 *	-char *filename: Nombre del fichero pcap a analizar
 *	
 * Salida:
 *	- 0 si no hay ningún problema
 *	- -1 si el nombre de fichero pasado es NULL o se produce algún error al procesar el fichero pcap especificado.
 *
 *  Autor: Javier Ramos de Santiago
 *
 *  (C) 2013 EPS-UAM 
 ****************************************************************************************************************************************************************/

int getSizeECDFPort(char* filename,char *filename_out,uint16_t port_src, uint16_t port_dst)
{
	
	/* PRACTICA: implementa esta funcion */
  	printf("%s/%d: 'getSizeECDF' no implementada...\n", __FILE__, __LINE__);
	return OK;

}
